/*
 * Load an ftc file and dump some info out.
 */

#if defined(_MSC_VER)
#include "err.h"
#else
#include <err.h>
#endif

#include <FreeImage.h>
#include <ATI_Compress.h>

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
/* #include <sys/stdint.h> */
#include <io.h>


#include "hexdump.h"
#include "bin_ops.h"

#define FTEX_FORMAT_DXT1 0
#define FTEX_FORMAT_RGB 1

void *feedback(float progress, void* foo1, void* foo2)
{
  int mipmap;

  mipmap = *(int *)foo1;
  printf("\r");
  printf("%.1f%% of mipmap %d", progress, mipmap);
  fflush(stdout);
  return NULL;
}

void FreeImageErrorHandler(FREE_IMAGE_FORMAT fif, const char *message) {
  printf("\n*** ");
  printf("%s Format\n", FreeImage_GetFormatFromFIF(fif));
  printf(message);
  printf(" ***\n");
}

int main (int argc, char **argv)
{
	char *file;
	int fd, tmp, ret, divisor;
	uint8_t *buffer, *b_tmp;
	off_t size;
	struct stat finfo;
	uint32_t version, width, height, i, j, mipmap_count, format_count, format, where, mipmap_size;

	if (argc != 2)
	{
		errx(EXIT_FAILURE, "usage: %s <file.ftc>", argv[0]);
	}

	file = argv[1];

	fd = open(file, O_RDONLY|O_BINARY, 0);

	if (fd < 0)
	{
		errx(EXIT_FAILURE, "%s", file);
	}

  FreeImage_SetOutputMessage(FreeImageErrorHandler);

	fstat(fd, &finfo);

	buffer = malloc(finfo.st_size);

	size = read(fd, buffer, finfo.st_size);

/*	printf("read: %d\n", size);*/

	if (memcmp(buffer, "FTEX", 4) != 0)
	{
		errx(EXIT_FAILURE, "File %s is not an FTEX file\n", file);
	}

	version = read_32(buffer + 4);
	width = read_32(buffer + 8);
	height = read_32(buffer + 12);
	mipmap_count = read_32(buffer + 16);
	format_count = read_32(buffer + 20);

	printf("%s: FTEX file version %d, %dx%d, %d mipmaps, %d formats, %d bytes\n",
		file, version, width, height, mipmap_count, format_count, size);

	/*pretty_print_block(buffer + 24, (format_count * 8));*/

	b_tmp = buffer + 24;
	for (i = 0 ; i < format_count ; i++)
	{
		format = read_32(b_tmp);
		where = read_32(b_tmp + 4);
		if (format == FTEX_FORMAT_DXT1)
			printf("dxt1 format\n");
		else if (format == FTEX_FORMAT_RGB)
		{
			printf("24bit RGB format\n");
			printf("Sorry, only .ftc supported here.\n");
			return 0;
		}
		else
			printf("unknown format: %d\n", format);

		tmp = where;
		for (j = 1; j < (mipmap_count + 1); j++)
		{
      char header[32];
      char *filename;
      int i2, mw, mh;
      FILE *outfd;
      ATI_TC_Texture src, dst;
      ATI_TC_ERROR ati_err;
      FIBITMAP *outbm;

			divisor = 2 << (j - 2);

			if (divisor == 0)
				divisor = 1;

      mw = width / divisor;
      mw = mw < 1 ? 1 : mw;

      mh = height / divisor;
      mh = mh < 1 ? 1 : mh;

			mipmap_size = read_32(buffer + tmp);
			printf("mipmapsize: %d level: %d %dx%d\n", mipmap_size, j, mw, mh);
//			pretty_print_block(buffer + tmp + 4, mipmap_size);

      filename = malloc(strlen(file) + 42);
      sprintf(filename, "%s_%02d_%dx%d.bmp%c", file, j, mw, mh, 0x00);

/*      if ((outfd = fopen(filename, "wb")) == NULL)
      {
        free(filename);
        errx(EXIT_FAILURE, "unable to open output file %s",
              filename);
      }*/

//      sprintf(header, "P6%c%d %d%c255%c%c", 0x0a, 4, 4, 0x0a, 0x0a, 0x00);
//      ret = fwrite(header, strlen(header), 1, outfd);

      src.dwSize = sizeof(src);
      src.dwWidth = mw;
      src.dwHeight = mh;
      src.dwPitch = 0;
      src.format = ATI_TC_FORMAT_DXT1;
      src.pData = buffer + tmp + 4;
      src.dwDataSize = ATI_TC_CalculateBufferSize(&src);

      printf("Compressed src: \n");
      pretty_print_block(src.pData, src.dwDataSize > 16 ? 16 : src.dwDataSize);

      printf("Ati: %d me: %d\n", src.dwDataSize, mipmap_size);

      dst.dwSize = sizeof(dst);
      dst.dwWidth = mw;
      dst.dwHeight = mh;
      dst.dwPitch = mw * 4;
      dst.format = ATI_TC_FORMAT_ARGB_8888;
      dst.dwDataSize = ATI_TC_CalculateBufferSize(&dst);
      dst.pData = (ATI_TC_BYTE*) malloc(dst.dwDataSize);

      printf("Uncompressed texture size is %d : %d\n", dst.dwDataSize, mw * mh * 4);

      ati_err = ATI_TC_ConvertTexture(&src, &dst, NULL, feedback, &j, NULL);
      printf("\n");

      if (ati_err != ATI_TC_OK)
      {
        errx(EXIT_FAILURE, "me: error while trying to compress texture: %d",
               ati_err);
      }

      printf("Decompressed: \n");
      pretty_print_block(dst.pData, dst.dwDataSize > 16 ? 16 : dst.dwDataSize);

      outbm = FreeImage_ConvertFromRawBits(dst.pData, mw, mh, mw * 4, 32,
        FI_RGBA_RED_MASK, FI_RGBA_GREEN_MASK, FI_RGBA_BLUE_MASK, 0);

      if (outbm == NULL)
      {
        errx(EXIT_FAILURE, "Failed to convert to a FIB\n");
      }

      if (FreeImage_IsTransparent(outbm))
      {
        printf("WARNING: %s : mipmap %d has transparent bits\n", file, j);
      }

      FreeImage_Save(FIF_BMP, outbm, filename, 0);

      free(filename);
      free(dst.pData);
//      fclose(outfd);
			tmp += mipmap_size + 4;
		}
	}

	close(fd);

	free(buffer);

	return 0;
}

